from .util import snake_case


class Volume(object):
    def __init__(self, image_name, path):
        base = 'slipway_' + snake_case(image_name)
        self.name = base + '_' + snake_case(path)
        self.path = path


class Volumes(object):
    def __init__(self, client, args, image):
        self.client = client
        self.args = args
        self.image = image

    def list(self):
        """
        Lists all volumes
        """
        return map(
            lambda volume: Volume(self.image.name, volume), self.image.volumes)

    def purge(self):
        """
        Clears all volumes tied to the given arguments
        """
        docker_volumes = self.client.volumes.list()
        for volume in self.list():
            found = None
            for docker_volume in docker_volumes:
                if volume.name == docker_volume.name:
                    found = docker_volume
            if found is not None:
                found.remove()

    def initialize(self):
        """
        Checks if the required volumer are present and creates them if missing.
        """
        docker_volumes = self.client.volumes.list()
        for volume in self.list():
            exist = False
            for docker_volume in docker_volumes:
                if volume.name == docker_volume.name:
                    exist = True
            if not exist:
                self.client.volumes.create(volume.name)



import sys
from argparse import ArgumentParser

import docker
from .container import Container


def parse_args():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='mode')
    subparsers.required = True

    start_parser = subparsers.add_parser('start')
    start_parser.add_argument('image')
    start_parser.add_argument('--volume', action='append')
    start_parser.add_argument('--environment', action='append')

    purge_parser = subparsers.add_parser('purge')
    purge_parser.add_argument('image')

    return parser.parse_args()


def main():
    if sys.version_info.major < 3:
        print('Python 2 is not supported')
        sys.exit(1)
    args = parse_args()
    client = docker.from_env()
    container = Container(client, args)
    if args.mode == 'start':
        container.image.initialize()
        container.volumes.initialize()
        container.binds.initialize()
        container.run()
    elif args.mode == 'purge':
        container.volumes.purge()




import os
from os import path
from .image import Image
from .volumes import Volumes
from .binds import Binds
from .util import snake_case


class Container(object):

    def __init__(self, client, args):
        self.client = client
        self.args = args
        self.image = Image(self.client, self.args)
        self.volumes = Volumes(self.client, self.args, self.image)
        self.binds = Binds(self.client, self.args, self.image)
        self.name = 'slipway_' + snake_case(self.args.image)

    def _entrypoint(self):
        script_dir = path.dirname(path.abspath(__file__))
        return path.join(script_dir, 'entrypoint.py')

    def _volumes_env(self):
        return ','.join(list(map(lambda vol: vol.path, self.volumes.list())))

    def _run_arguments(self):
        arguments = [
            'docker',
            'run',
            '--net=host',
            '--user', 'root',
            '--rm',
            '-ti',
            '--detach-keys', 'ctrl-q,ctrl-q',
            '--name', self.name,
            '-e', 'GH_USER',
            '-e', 'GH_PASS',
            '-e', 'DISPLAY',
            '-e', 'SLIPWAY_USER={}'.format(self.image.user),
            '-e', 'SLIPWAY_VOLUMES={}'.format(self._volumes_env()),
            '-v', self._entrypoint() + ':/slipway-entrypoint.py:ro',
            '--entrypoint', '/slipway-entrypoint.py'
        ]

        for env in self.args.environment or []:
            arguments.append('-e')
            arguments.append(env)

        for bind in self.binds.list():
            arguments.append('-v')
            argument = bind.host_path + ':' + bind.container_path
            if 'ro' in bind.type:
                argument += ':ro'
            arguments.append(argument)

        for volume in self.volumes.list():
            arguments.append('--mount')
            arguments.append(
                'source={},target={}'.format(volume.name, volume.path))

        arguments.append(self.image.name)
        arguments.append('tmux')
        arguments.append('new')
        return arguments

    def run(self):
        os.execvp('docker', self._run_arguments())



import pwd
import os
from os import environ, path


class Bind(object):
    def __init__(self, host_path, container_path, type='d'):
        self.host_path = host_path
        self.container_path = container_path
        self.type = type


class Binds(object):
    """
    A bind is a type of volume which maps a directory or file from the host
    into the container.
    """

    def __init__(self, client, args, image):
        self.client = client
        self.args = args
        self.image = image

    def initialize(self):
        """
        Creates the directory with appropriate permissions if it doesn't
        exist. Otherwise if we let docker do this the permissions will be
        set as root, which we don't want.
        """
        for bind in self.list():
            if not path.exists(bind.host_path):
                if 'd' in bind.type:
                    os.makedirs(bind.host_path)
                elif 'f' in bind.type:
                    open(bind.host_path, 'w+').close()

    def list(self):
        """
        List all bind mounts, including the default ones.
        """
        home_mappings = [
            ('d', 'workspaces'),
            ('d', '.ssh'),
            ('f', '.gitconfig')
        ]
        for type, mapping in home_mappings:
            host_path = path.join(environ['HOME'], mapping)
            container_path = path.join(self.image.home, mapping)
            yield Bind(host_path, container_path, type)

        if path.isfile('/etc/localtime'):
            yield Bind('/etc/localtime', '/etc/localtime', 'fro')
        if path.exists('/tmp/.X11-unix'):
            yield Bind('/tmp/.X11-unix', '/tmp/.X11-unix', 'fro')

        uid = pwd.getpwnam(environ['USER']).pw_uid
        gnupg_socket_path = path.join('/run/user', str(uid), 'gnupg')
        if path.exists(gnupg_socket_path):
            # FIXME: how to get uid of user in container?
            yield Bind(gnupg_socket_path, '/run/user/1000/gnupg', 'dro')

        gnupg_path = path.join(environ['HOME'], '.gnupg')
        if path.exists(gnupg_path):
            container_path = path.join(self.image.home, '.gnupg')
            yield Bind(gnupg_path, container_path, 'd')

        for volume in self.args.volume or []:
            parts = volume.split(':')
            host_path = parts[0]
            container_path = parts[1]
            type = 'd'
            if path.isfile(host_path):
                type = 'f'
            if len(parts) == 3 and parts[2] == 'ro':
                type += 'ro'
            yield Bind(host_path, container_path, type)



from docker.errors import ImageNotFound
from os import path


class Image(object):
    def __init__(self, client, args):
        self.client = client
        self.args = args
        self.name = args.image
        self._image = None

    def _docker_image(self):
        if self._image is None:
            self._image = self.client.images.get(self.name)
        return self._image

    def initialize(self):
        """
        Pulls the image if not present
        """
        try:
            self._docker_image()
        except ImageNotFound:
            message = 'Image {} not found, attempting to pull down'
            print(message.format(self.name))
            self.client.images.pull(self.name)

    @property
    def volumes(self):
        """
        Returns all volumes tied to the image's container configuration
        """
        config = self._docker_image().attrs['ContainerConfig']
        if config['Volumes'] is None:
            return []
        return list(config['Volumes'].keys())

    @property
    def entrypoint(self):
        return self._docker_image().attrs['ContainerConfig']['Entrypoint']

    @property
    def user(self):
        return self._docker_image().attrs['ContainerConfig']['User']

    @property
    def home(self):
        return path.join('/home', self.user)



from setuptools import setup, find_packages

setup(
    name = 'slipway',
    packages = find_packages(),
    version = '0.4.0',
    description = 'CLI tool for managing development containers',
    author = 'Jonathan Boudreau',
    author_email = 'jonathan.boudreau.92@gmail.com',
    url = 'https://github.com/AGhost-7/slipway',
    download_url = '',
    keywords = ['docker', 'development'],
    classifiers = [],
    install_requires = [
        'docker >= 3.0.0'
    ],
    entry_points = { 
        'console_scripts': [
            'slipway=slipway:main'
        ]
    }
)



#!/usr/bin/env python3

"""
Entrypoint script
"""

import os
from os import environ
import sys
import pwd


entrypoint = environ.get('SLIPWAY_ENTRYPOINT')
user_name = environ.get('SLIPWAY_USER')

# All we need to do is correct the permisions in the volumes and call it a day.
user = pwd.getpwnam(user_name)
uid = user.pw_uid
gid = user.pw_gid

for volume in environ.get('SLIPWAY_VOLUMES').split(','):
    if len(volume) > 0:
        os.chown(volume, uid, gid)

# Essentially, doing this but in python:
# exec su -c "$@" $SLIPWAY_USER

args = ['sudo', '--preserve-env', '-u', user_name]

if entrypoint:
    args.append(entrypoint)

args.extend(sys.argv[1:])

os.execvp('sudo', args)
